/*!
 \file  basetypes.h
 \brief This file contains base type definitions

 \author David C. Anastasiu
 */

#ifndef _L2KNNG_BASETYPES_H_
#define _L2KNNG_BASETYPES_H_



/* Index type should be a signed int type */
#ifdef IDXTYPE_LONG
#define PRNT_IDXTYPE "%ld"
#define IDX_MAX INT64_MAX
#define IDX_MIN INT64_MIN
typedef int64_t idx_t;
#elif defined IDXTYPE_SHORT
#define IDX_MAX SHRT_MAX
#define IDX_MIN SHRT_MIN
#define PRNT_IDXTYPE "%hi"
typedef short idx_t;
#else
#define IDXTYPE_INT
#define IDX_MAX INT32_MAX
#define IDX_MIN INT32_MIN
#define PRNT_IDXTYPE "%d"
typedef int32_t idx_t;
#endif

/* Pointer type should be a signed int type */
#ifdef PTRTYPE_INT
#define PRNT_PTRTYPE "%d"
#define PTR_MAX INT32_MAX
#define PTR_MIN INT32_MIN
typedef int32_t ptr_t;
#elif defined PTRTYPE_LONG
#define PRNT_PTRTYPE "%ld"
#define PTR_MAX INT64_MAX
#define PTR_MIN INT64_MIN
typedef int64_t ptr_t;
#else
#define PTRTYPE_SSZIE
#define PTR_MAX INT64_MAX
#define PTR_MIN INT64_MIN
#define PRNT_PTRTYPE "%zu"
typedef ssize_t ptr_t;
#endif


/* Value type should be a float or double */
#ifdef VALTYPE_DOUBLE
typedef double val_t;
#define VAL_MAX DBL_MAX
#define VAL_MIN DBL_MIN
#define PRNT_VALTYPE "%g"
#else
#define VALTYPE_FLOAT
#define VAL_MAX FLT_MAX
#define VAL_MIN FLT_MIN
#define PRNT_VALTYPE "%f"
typedef float val_t;
#endif

/* Accumulator type should be a float or double */
#ifdef ACCUMTYPE_DOUBLE
typedef double accum_t;
#define ACCUM_MAX DBL_MAX
#define ACCUM_MIN DBL_MIN
#define PRNT_ACCUMTYPE "%g"
#else
#define ACCUMTYPE_FLOAT
#define ACCUM_MAX FLT_MAX
#define ACCUM_MIN FLT_MIN
#define PRNT_ACCUMTYPE "%f"
typedef float accum_t;
#endif

#define MAXIDX	(1<<8*sizeof(idx_t)-2)

typedef unsigned int uint;

#endif
