

#ifndef DA_L2KNN_PL2KNN
#define DA_L2KNN_PL2KNN

#include "includes.h"


/** Definitions */
#define PL2KNN_QHASH_DENSE 1
#define PL2KNN_QHASH_HASH  2
#ifndef PL2KNN_QHASH
#define PL2KNN_QHASH PL2KNN_QHASH_DENSE
#endif
#ifndef PL2KNN_QHASH_HTSIZE
#define PL2KNN_QHASH_HTSIZE 0
#endif

/* run k-select on candidates before entering into local heap */
#ifndef PL2KNNA_PRESELECT
#define PL2KNNA_PRESELECT false
#endif

/* use pruning when computing dot-products in the approximate algo */
#ifndef PL2KNNA_PRUNE
#define PL2KNNA_PRUNE false
#endif
#ifndef PL2KNNA_PRUNE_NCACHE
#define PL2KNNA_PRUNE_NCACHE true
#endif
#ifndef PL2KNNA_PRUNE_MIN
#define PL2KNNA_PRUNE_MIN 0.5
#endif
#ifndef PL2KNNA_PRUNE_SZ
#define PL2KNNA_PRUNE_SZ 20	/* set to high number to test all features */
#endif

/* execute local heap inserts immediately -- only if PL2KNNA_PRUNE is true */
#ifndef PL2KNNA_HINSERT
#define PL2KNNA_HINSERT false
#endif

#ifndef PL2KNN_DYNAMIC
#define PL2KNN_DYNAMIC 10
#endif

#ifndef PL2KNN_DYNAMIC_FAST
#define PL2KNN_DYNAMIC_FAST 128
#endif

#define L2KNN_CINIT -1
#define L2KNN_CPRUNED -2
#define L2KNN_CPRUNED_NBRS -3
#define L2KNN_CPRUNED_GRAPH -4


void pl2knnPrintCompileOptions ();

ptr_t pl2knnChooseInitialNeighbors (const da_csr_t * const docs,
				    da_tsearch_t * const ts,
				    da_iakv_t * const qcands,
				    idx_t * const cnt, const idx_t rid);

void pl2knnFindInitialNeighbors (params_t * const params,
				 da_csr_t * const docs,
				 da_csr_t * const fdocs,
				 da_csr_t * const bnbrs,
				 idx_t * const nnbrs,
				 val_t * const rlen,
				 da_iapq_t * const *const knng,
				 da_iakv_t * const *const qnn,
				 idx_t * const *const cnts,
				 da_tsearch_t * const *const search);

size_t pl2knnEnhanceNeighborhood (params_t * const params,
				  da_csr_t * restrict const docs,
				  da_csr_t * restrict const fdocs,
				  da_csr_t * restrict const bnbrs,
				  idx_t * restrict const nnbrs,
				  val_t const *const rlen,
				  da_iapq_t * const *const knng,
				  da_iakv_t * const *const qnn,
				  idx_t * const *const cnts,
				  da_tsearch_t * const *const search);

accum_t psddotp (size_t const sz,
		 idx_t const *const ind,
		 val_t const *const val, val_t const *const qrow);

accum_t psdpdotp (val_t const msim,
		  size_t const sz,
		  idx_t const *restrict const ind,
		  val_t const *restrict const val,
		  val_t const *restrict const qrow);

void pl2knnTransferCounts (params_t * const params,
			   da_tsearch_t * const *const search);

void da_pl2knnSortTopKValues (da_csr_t * const mat,
			      char const what,
			      idx_t const topk,
			      idx_t const offset,
			      da_tsearch_t * const *const search);

void pl2knnReorderCols (da_csr_t * const docs, char const order);

size_t pl2knnCheckCandidates (idx_t const qid,
			      idx_t const start,
			      bool const checkfull,
			      ptr_t const *restrict const ptr,
			      idx_t const *restrict const ind,
			      val_t const *restrict const val,
			      da_iapq_t * restrict const lnn,
			      da_tsearch_t * restrict const ts,
			      size_t const ncands,
			      da_iakv_t * restrict const anbrs,
			      idx_t * const *restrict const cnts);

idx_t pl2knnIndexP (idx_t const nsims,
		    da_iakv_t * const msim,
		    params_t * const params,
		    da_iapq_t * const *const knng,
		    da_csr_t const *const docs,
		    ptr_t * const pszs, da_csr_t * const idx);

void pl2KnnFindQueryNeighbors (da_tsearch_t * const ts, idx_t const rid,	/* global query row id in original doc order */
			       idx_t const dqid,	/* local id in db tile order -- row id in idx */
			       idx_t const lqid,	/* local id in query tile order -- row id in bnbrs */
			       val_t const rct,	/* minimum neighborhood similarity among indexed objects */
			       da_csr_t const *const docs,	/* input documents */
			       da_csr_t const *const idx,	/* db tile index */
			       da_csr_t const *const bnbrs,	/* current version of tile neighborhoods */
			       da_iapq_t * const *const knng,	/* the output graph */
			       da_iakv_t * const qnn,	/* candidate update list for query row */
			       idx_t * const qcnt,	/* counts array for query row */
			       da_iakv_t * const msim);	/* minimum neighborhood similarities and current processing order */


ptr_t pl2knnChooseInitialNeighbors2 (const da_csr_t * const docs,
				     accum_t * accum,
				     da_iakv_t * anbrs,
				     idx_t * cnt,
				     const idx_t nthr,
				     const idx_t nrows,
				     const idx_t bsz,
				     const idx_t nn, const idx_t rid);

void pl2knnFindInitialNeighbors2 (params_t * const params,
				  da_csr_t * const docs,
				  idx_t * const nnbrs,
				  da_iapq_t ** knng,
				  accum_t ** accums,
				  val_t ** qrows,
				  da_iakv_t ** qnn,
				  da_iakv_t ** qtmp,
				  idx_t ** cnts, ptr_t ** rlocs);

size_t pl2knnEnhanceNeighborhood2 (params_t * const params,
				   da_csr_t * const docs,
				   idx_t * const nnbrs,
				   da_iapq_t ** knng,
				   val_t ** qrows,
				   da_iakv_t ** qnn,
				   da_iakv_t ** qtmp,
				   idx_t ** cnts,
				   ptr_t ** rlocs, char **tags);

accum_t psddotp2 (size_t sz, idx_t * ind, val_t * val, val_t * qrow);

void init_tsearch (size_t const tid,
		   idx_t const nrows,
		   idx_t const ncols,
		   idx_t const mrl,
		   params_t const *const params, da_tsearch_t * const ts);

void free_tsearch (da_tsearch_t ** tsp);

void free_all_tsearch (da_tsearch_t *** searchp, idx_t const sz);

accum_t psdpdotp2 (val_t const msim,
		   size_t const sz,
		   idx_t const *restrict const ind,
		   val_t const *restrict const val,
		   val_t const *restrict const len,
		   val_t const *restrict const qrow,
		   val_t const *restrict const qrln);

idx_t da_GetKSimilarRows (da_csr_t * mat, idx_t rid, char noSelfSim,
                    idx_t nqterms, idx_t * qind, val_t * qval, char simtype,
                    idx_t nsim, da_ivkv_t * hits, idx_t * i_marker,
                    da_ivkv_t * i_cand, idx_t * ncands);

void l2knnReorderCols (da_csr_t * docs, idx_t * csizes, idx_t * cperm);

void l2apReorderCols (da_csr_t * docs, idx_t * csizes, idx_t * cperm);


#endif
