/*
 * sdhash.h
 *
 *  Created on: Apr 17, 2016
 *      Author: David C. Anastasiu
 */

#ifndef DA_L2KNN_SDHASH_H_
#define DA_L2KNN_SDHASH_H_

#include "includes.h"

/* default load factor */
#ifndef SDHASH_LF
#define SDHASH_LF 3.75
#endif



#define sdhashHash(key, htmask) \
    (key & htmask)

/**
 * Create an sdhash table. Set htsize to 0 to automatically determine based on maxfid and maxnnz.
 * \param maxsz Total number of elements that will ever be added to the hash at the same time
 * \param htsize The size of the hash table (must be a power of 2)
 * \param maxfid The max size of an id that will be looked up in the table (used to estimate htsize)
 * NOTE: UNSAFE, no bounds checking for maxsz. Ensure at most maxsz elements will be set into the hash at a time.
 */
sdhash *sdhashCreate (size_t const maxsz,
		      size_t const htsize, size_t const maxfid);

/**
 * Free an sdhash structure
 */
void sdhashFree (sdhash ** const pt);

/**
 * Print the current contents of an sdhash structure
 */
void sdhashPrint (sdhash const *const t);


/**
 * Set the query values in the hash table
 * NOTE: UNSAFE, no bounds checking for maxsz. Ensure at most maxsz elements will be set into the hash at a time.
 */
void sdhashSet (sdhash * const t,
		idx_t const *const qind,
		val_t const *const qval, ptr_t const qsz);

/**
 * Set the query values in the hash table
 * NOTE: UNSAFE, no bounds checking for maxsz. Ensure at most maxsz elements will be set into the hash at a time.
 * NOTE2: Prefer sdhashSet over this method, as it has to sort overflow every time an item is added to overflow.
 */
void sdhashSetKeyVal (sdhash * const t, idx_t const key, val_t const val);

/**
 * Remove the query values from the hash table
 */
void sdhashUnset (sdhash * const t, idx_t const *const qind, ptr_t const qsz);

/**
 * Compute the dot-product between the sparse vector stored in the hash table and another sparse vector
 */
accum_t sdhashDotp (sdhash * const t,
		    idx_t const *const ind,
		    val_t const *const val, ptr_t const sz);


val_t sdhashVal (sdhash const *const t, idx_t const key);

bool sdhashIsset (sdhash const *const t, idx_t const key);


#endif /* DA_L2KNN_SDHASH_H_ */
