/*!
\file  bkheap.c
\brief This file implements various tests for the data structures in this program
*/
#include "bkheap.h"
#include "sdhash.h"



#define TEST_INFO(hid, action, key, val) {printf("h%d: %s %d %g\n", hid, action, key, (double)val); fflush(stdout);}


void testBkheap ();
void testSdhash (bool const prnt);



#define TBKH_O 127
#define TBKH_K   13
#define TBKH_CNT 30000
void
testBkheap ()
{
  ptr_t n, i, j, k, z1, z2, z3, z4;
  double p;
  bkheap *h1, *h2, *h3;
  idx_t *nums;


  n = TBKH_O + TBKH_K;
  h1 = bkheapCreate (TBKH_K, TBKH_O, 1);
  h2 = bkheapCreate (TBKH_K, TBKH_O, 1);
  h3 = bkheapCreate (TBKH_K, TBKH_O, 1);
  h1->duplicates = 0;
  h2->duplicates = 0;
  h3->duplicates = 0;
  nums = da_imalloc (n, "testBkheap: nums");
  da_irandArrayPermuteFine (n, nums, 1);

  printf ("heaps created\n");
  fflush (stdout);

  for (i = 0; i < TBKH_K; ++i)
    {
      bkheapInsert (h1, nums[i], nums[i]);
      TEST_ASSERT (bkheapCheck (h1));
      bkheapInsertHeapNew (h2, nums[i], nums[i]);
      TEST_ASSERT (bkheapCheck (h2));
      TEST_ASSERT (bkheapMin (h1) == bkheapMin (h2));
      h3->data[i].key = nums[i];
      h3->data[i].val = nums[i];
    }

  h3->sz = TBKH_K;
  TEST_INFO (3, "buildHeap", -1, -1);
  bkheapBuildHeap (h3);
  TEST_ASSERT (bkheapCheck (h3));
  TEST_ASSERT (bkheapMin (h3) == bkheapMin (h2));

  printf ("inserted k items\n");
  fflush (stdout);

  for (i = 0; i < TBKH_K; ++i)
    {
      TEST_ASSERT (bkheapExists (h1, nums[i]));
      TEST_ASSERT (bkheapExists (h2, nums[i]));
      TEST_ASSERT (bkheapExists (h3, nums[i]));
    }

  printf ("existence check passed \n");
  fflush (stdout);

  for (; i < n; ++i)
    {
      bkheapInsert (h1, nums[i], nums[i]);
      bkheapCheck (h1);
      bkheapInsertHeap (h2, nums[i], nums[i]);
      bkheapCheck (h2);
      bkheapInsertHeapNew (h3, nums[i], nums[i]);
      bkheapCheck (h3);
      TEST_ASSERT (bkheapMin (h3) == bkheapMin (h2));
    }

  printf ("inserted remaining items\n");
  fflush (stdout);

  TEST_INFO (1, "compact", -1, -1);
  bkheapCompact (h1);
  bkheapCheck (h1);
  TEST_ASSERT (bkheapMin (h1) == bkheapMin (h2));

  printf ("h1 compact succeeded\n");
  fflush (stdout);

  bkheapClear (h2);
  bkheapReset (h3);
  for (i = 0; i < n; ++i)
    {
      bkheapInsert (h2, i + 1, i + 1);
    }

  printf ("heaps reset and h2 populated in order\n");
  fflush (stdout);

  TEST_INFO (2, "copy to h3", -1, -1);
  bkheapCopy (h2, h3);
  TEST_ASSERT (bkheapMin (h2) == 1);
  TEST_ASSERT (bkheapMin (h3) == bkheapMin (h2));

  TEST_INFO (2, "insert", (idx_t) n + 1, (double) n + 1);
  bkheapInsert (h2, n + 1, n + 1);	/* should trigger compact */

  TEST_ASSERT (bkheapMin (h2) == n + 1 - h2->maxsz);
  TEST_INFO (3, "compact", -1, -1);
  bkheapCompact (h3);

  TEST_ASSERT (bkheapMin (h3) == n + 1 - h2->maxsz);
  TEST_INFO (3, "insert", (idx_t) n + 1, (double) n + 1);
  bkheapInsert (h3, n + 1, n + 1);
  TEST_ASSERT (bkheapMin (h3) == n + 1 - h2->maxsz);
  bkheapCompact (h3);
  TEST_ASSERT (bkheapMin (h3) == n + 2 - h2->maxsz);

  printf ("0 element insert and trigger successful \n");
  printf ("clearing heaps and inserting random runs of random values \n");
  fflush (stdout);

  h1->duplicates = 1;
  h2->duplicates = 1;
  h3->duplicates = 1;

#ifdef EXTRACOUNTS
  z1 = z2 = p = 0;
#endif
  z3 = z4 = 0;
  for (i = 0; i < TBKH_CNT; ++i)
    {
      n = da_irandInRange (10000);
      bkheapClear (h1);
      bkheapClear (h2);
      for (j = 0; j < n; ++j)
	{
	  k = da_irandInRange (1e+6);
	  bkheapInsert (h1, (idx_t) k, (double) k);
	  bkheapInsertHeap (h2, (idx_t) k, (double) k);
	}
      bkheapCheck (h1);
      bkheapCheck (h2);
      z3 += h1->hz;
      z4 += h2->hz;

      bkheapCompact (h1);
      if (da_abs (bkheapMin (h1) - bkheapMin (h2)) > 1e-6)
	{
	  printf ("\n   h1 min != h2 min\n\n");
	  printf ("\n   h1:\n");
	  bkheapPrint (h1);
	  printf ("\n   h2:\n");
	  bkheapResetLocator (h2);
	  bkheapPrint (h2);

	  da_ivkvsorti (h1->sz, h1->data);
	  da_ivkvsorti (h2->sz, h2->data);
	  bkheapPrint (h1);
	  bkheapPrint (h2);

	  da_errexit ("h1 min != h2 min");
	}
      da_ivkvsorti (h1->maxsz, h1->data);
      da_ivkvsorti (h2->maxsz, h2->data);
      for (j = 0; j < h1->maxsz; ++j)
	{
	  TEST_ASSERT (h1->data[j].key == h2->data[j].key);
	  TEST_ASSERT (da_abs (h1->data[j].val - h2->data[j].val) < 1e-6);
	}
#ifdef EXTRACOUNTS
      z1 += h1->hc;
      z2 += h2->hc;
      p += h1->hc / ((double) h2->hc);
    }
  p /= TBKH_CNT;
  printf ("Collision ratio: %f\n", z1 / ((double) z2));
  printf ("Avg. collision ratio: %f\n", p);
#else
    }
#endif
  printf ("Avg. locator overflow length h1: %.17g\n", z3 / (double) TBKH_CNT);
  printf ("Avg. locator overflow length h2: %.17g\n", z4 / (double) TBKH_CNT);

  printf ("\nRandom inserts successful. Test done.\n\n");
  fflush (stdout);

  bkheapFree (&h1);
  bkheapFree (&h2);
  bkheapFree (&h3);
  da_free ((void **) &nums, LTERM);
}

#define TSDH_N1 127
#define TSDH_N2 135
#define TSDH_MFID 100000
void
testSdhash (bool const prnt)
{

  size_t i, j;
  idx_t k;
  idx_t ind1[TSDH_N1];		/* v1 */
  val_t val1[TSDH_N1];
  idx_t ind2[TSDH_N2];		/* v2 */
  val_t val2[TSDH_N2];
  double s, s1, s2;
  sdhash *t;
  bit *c;

  ba_init ();
  t = sdhashCreate (da_max (TSDH_N1, TSDH_N2), 0, TSDH_MFID);	/* create table structure */
  c = ba_new (TSDH_MFID + 1);
  for (i = 0; i < TSDH_N1; ++i)
    {				/* random v1 */
      k = da_irandInRange (TSDH_MFID);
      while (ba_value (c, k))
	{
	  k = da_irandInRange (TSDH_MFID);
	}
      ba_assign_true (c, k);
      ind1[i] = k;
      val1[i] = da_irandInRange (10 * TSDH_MFID) / ((double) TSDH_MFID);
    }
  s = da_vnorm2 (TSDH_N1, val1, 1);	/* normalize v1 */
  for (i = 0; i < TSDH_N1; ++i)
    {
      val1[i] /= s;
    }


  ba_all_assign (c, TSDH_MFID + 1, false);
  for (i = 0; i < TSDH_N2; ++i)
    {				/* random v2 */
      k = da_irandInRange (TSDH_MFID);
      while (ba_value (c, k))
	{
	  k = da_irandInRange (TSDH_MFID);
	}
      ba_assign_true (c, k);
      ind2[i] = k;
      val2[i] = da_irandInRange (17 * TSDH_MFID) / ((float) TSDH_MFID);
    }
  s = da_vnorm2 (TSDH_N2, val2, 1);	/* normalize v2 */
  for (i = 0; i < TSDH_N2; ++i)
    {
      val2[i] /= s;
    }

  /* Set v1 in the hash table */
  sdhashSet (t, ind1, val1, TSDH_N1);
  if (prnt)
    {
      printf ("v1 set:\n");
      sdhashPrint (t);
    }

  /* Self-similarity v1, result should be close to 1 */
  s = sdhashDotp (t, ind1, val1, TSDH_N1);
  if (prnt)
    {
      printf ("v1 self-dotp: %f\n", da_vdot (TSDH_N1, val1, 1, val1, 1));
      printf ("v1 self-similarity: %f\n", s);
    }
  TEST_ASSERT (da_abs (s - 1.0) < 1e-4);

  /* compute similarity with v2 */
  s1 = sdhashDotp (t, ind2, val2, TSDH_N2);

  /* clear hash table */
  sdhashUnset (t, ind1, TSDH_N1);
  TEST_ASSERT (t->sz == 0);
  if (prnt)
    {
      printf ("v1 unset:\n");
      sdhashPrint (t);
    }

  /* Set v2 in the hash table */
  sdhashSet (t, ind2, val2, TSDH_N2);
  if (prnt)
    {
      printf ("v1 set:\n");
      sdhashPrint (t);
    }

  /* Self-similarity v2, result should be close to 1 */
  s = sdhashDotp (t, ind2, val2, TSDH_N2);
  if (prnt)
    {
      printf ("v2 self-dotp: %f\n", da_vdot (TSDH_N2, val2, 1, val2, 1));
      printf ("v2 self-similarity: %f\n", s);
    }
  TEST_ASSERT (da_abs (s - 1.0) < 1e-4);

  /* compute similarity with v2 */
  s2 = sdhashDotp (t, ind1, val1, TSDH_N1);
  if (prnt)
    {
      printf ("sim(v1,v2) = %f, sim(v2,v2) = %f\n", s1, s2);
    }

  TEST_ASSERT ((da_abs (s1 - s2) < 1e-4));

  /* clear hash table */
  sdhashUnset (t, ind2, TSDH_N2);
  TEST_ASSERT (t->sz == 0);
  if (prnt)
    {
      printf ("v1 unset:\n");
      sdhashPrint (t);
    }

  sdhashFree (&t);
  da_free ((void **) &c, LTERM);

}

/**
 * Main entry point to L2KNNG.
 */
void
test (params_t * params)
{


  size_t i;

  timer_start (params->timer_1);


//    testBkheap();

  for (i = 0; i < 1e+5; ++i)
    {
      testSdhash (false);
    }
  printf ("Burn-in test repeated %d times...\n", (int) 1e+5);
  testSdhash (true);


  timer_stop (params->timer_1);
  da_printTimerLong ("Test time: ", params->timer_1);
}
